package auth

import (
	"context"
	"errors"
	"net/http"

	"github.com/go-kit/kit/log/level"

	"git.c3pb.de/gbe/invinoveritas/log"
)

type contextKey string

var ErrAuthFailed = errors.New("authentication failed")

// User is an application user.
type User struct {
	Name    string
	IsAdmin bool
}

func (u User) String() string {
	return "{name:" + u.Name + "}"
}

type Provider interface {
	// Valid checks whether the given token represents a valid user session. If that is the case, the user
	// is returned. Otherwise, an error is returned.
	Valid(ctx context.Context, token string) (*User, error)
}

// Require wraps hdlr so that it requires authentication to use. Requests handled by hdlr will have the user
// name attached to their context. Use the User function to retrieve it.
//
// If authentication fails, for example because the users' session is no longer valid or doesn't exist, the
// request will be handled by the authFailed handler. It will receive the original request and response writer.
// This is triggered by provider returning an error wrapping ErrAuthFailed from its `Valid` method.
func Require(next http.Handler, authFailed http.Handler, provider Provider) http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		cookie, err := r.Cookie("session")
		if errors.Is(err, http.ErrNoCookie) {
			level.Error(log.Get(r)).
				Log("error", err,
					"msg", "auth failed")

			authFailed.ServeHTTP(w, r)
			return
		}

		if err != nil {
			// We shouldn't end up in here, but better safe than sorry.
			level.Error(log.Get(r)).
				Log("error", err,
					"msg", "Can't get auth cookie")

			http.Error(w, err.Error(), http.StatusInternalServerError)
			return
		}

		level.Debug(log.Get(r)).
			Log("cookie", cookie,
				"msg", "got auth cookie")

		user, err := provider.Valid(r.Context(), cookie.Value)
		if errors.Is(err, ErrAuthFailed) {
			// User is not authenticated: session expired or doesn't exist
			level.Error(log.Get(r)).
				Log("error", err,
					"msg", "auth failed")

			authFailed.ServeHTTP(w, r)
			return
		}

		if err != nil || user == nil {
			if user == nil && err == nil {
				// The auth provider did something weird: no error, but no user either
				err = errors.New("no user")
			}

			// Some error that isn't because the session is expired or doesn't exist (DB?)
			level.Error(log.Get(r)).
				Log("error", err,
					"msg", "can't authenticate user")

			http.Error(w, err.Error(), http.StatusInternalServerError)
			return
		}

		level.Info(log.Get(r)).
			Log("user", *user,
				"msg", "user authenticated")

		key := contextKey("user")
		ctx := context.WithValue(r.Context(), key, *user)

		next.ServeHTTP(w, r.WithContext(ctx))
	})
}

// Get returns the user from r's context. If there is no user in r's context, returns an empty user.
func Get(r *http.Request) User {
	val := r.Context().Value(contextKey("user"))
	if val == nil {
		return User{}
	}

	return val.(User)
}
